import functools
import json
from pkg_resources import resource_filename

from ...errors import FileFormatMetaDataError


__all__ = ["ReadJPKMetaKeyError",
           "get_primary_meta_recipe",
           "get_secondary_meta_recipe",
           ]


class ReadJPKMetaKeyError(FileFormatMetaDataError):
    pass


@functools.lru_cache()
def get_primary_meta_recipe():
    with open(resource_filename("afmformats.formats.fmt_jpk",
                                "jpk_meta_primary.json")) as fd:
        meta_recipe_pri = json.load(fd)
    return meta_recipe_pri


@functools.lru_cache()
def get_secondary_meta_recipe():
    with open(resource_filename("afmformats.formats.fmt_jpk",
                                "jpk_meta_secondary.json")) as fd:
        meta_recipe_sec = json.load(fd)
    return meta_recipe_sec



import pathlib

import numpy as np

import pytest

import afmformats

data_path = pathlib.Path(__file__).resolve().parent / "data"


@pytest.mark.parametrize("name,size,meta", [
    ("fmt-jpk-fd_spot3-0192.jpk-force", 4000, {}),
    ("fmt-afm-workshop-fd_single_2020-02-14_13.41.25.csv", 2030, {}),
    ("fmt-ntmdt-txt-fd_2015_01_17_gel4-0,1_mQ_adh_6B_Curve_DFL_Height_51.txt",
     1644,
     {"sensitivity": 61, "spring constant": 0.055}),
])
def test_length(name, size, meta):
    jpkfile = data_path / name
    fdat = afmformats.load_data(jpkfile,
                                modality="force-distance",
                                meta_override=meta,
                                )[0]
    assert len(fdat) == size
    assert np.all(fdat["index"] == np.arange(size))


def test_repr_str():
    jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
    fd = afmformats.load_data(jpkfile)[0]
    assert "AFMForceDistance" in str(fd)
    assert "fmt-jpk-fd_spot3-0192.jpk-force" in str(fd)
    assert "AFMForceDistance" in repr(fd)
    assert "fmt-jpk-fd_spot3-0192.jpk-force" in repr(fd)


def test_reset():
    jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
    fd = afmformats.load_data(jpkfile)[0]
    fd["tip position"] = np.linspace(0, 1, len(fd))
    assert "tip position" in fd
    assert "tip position" not in fd.columns_innate
    fd.reset_data()
    assert "tip position" not in fd


def test_segment():
    jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
    fd = afmformats.load_data(jpkfile)[0]
    assert np.sum(fd["segment"] == 0) == 2000
    assert fd.appr["segment"].size == 2000

    new_segment = np.zeros_like(fd["segment"])
    new_segment[500:] = 1
    fd["segment"] = new_segment
    assert np.sum(fd["segment"] == 0) == 500
    assert fd.appr["segment"].size == 500

    new_segment2 = np.zeros_like(fd["segment"])
    new_segment2[600:] = 1
    fd["segment"] = new_segment2
    assert np.sum(fd["segment"] == 0) == 600
    assert fd.appr["segment"].size == 600


def test_segment_set_item():
    jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
    fd = afmformats.load_data(jpkfile)[0]

    fd["tip position"] = np.ones(len(fd))
    assert np.all(fd.appr["tip position"] == np.ones(2000))
    fd.appr["tip position"] = 2*np.ones(2000)
    assert np.all(fd.appr["tip position"] == 2*np.ones(2000))
    assert np.all(fd.retr["tip position"] == np.ones(2000))
    assert np.all(fd["tip position"] == np.concatenate([2*np.ones(2000),
                                                        np.ones(2000)]))


def test_segment_set_item_2():
    jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
    fd = afmformats.load_data(jpkfile)[0]

    data = fd["height (measured)"]
    fd.appr["height (measured)"] = np.arange(2000)

    # test override not taking place
    assert not np.all(data[:2000] == np.arange(2000))
    # data access tests
    assert np.all(data[2000:] == fd["height (measured)"][2000:])
    assert np.all(fd["height (measured)"][2000:]
                  == fd.retr["height (measured)"])
    # test with private property
    assert np.all(fd.appr["height (measured)"]
                  == fd.appr._data["height (measured)"][:2000])
    assert not np.all(fd.appr["height (measured)"]
                      == fd.appr._raw_data["height (measured)"][:2000])



"""MetaData class"""
import copy

import numpy as np

import afmformats.meta as am


def test_parse_time():
    assert am.parse_time("16:10:20") == "16:10:20"
    assert am.parse_time("6:1:2") == "06:01:02"
    assert am.parse_time("6:1:2 PM") == "18:01:02"
    assert am.parse_time("6:1:2.0 PM") == "18:01:02"
    assert am.parse_time("6:1:2.001 PM") == "18:01:02.001"
    assert am.parse_time("6:1:2.0010 PM") == "18:01:02.001"


def test_get_ids():
    md = am.MetaData({"date": "2020-04-01",
                      "time": "21:56:30",
                      "enum": "2"})
    assert md["session id"] == "2020-04-01_21:56:30"
    assert md["curve id"] == "2020-04-01_21:56:30_2"

    # override session
    md["session id"] = "peter"
    assert md["session id"] == "peter"
    assert md["curve id"] == "peter_2"

    # override curve
    md["curve id"] = "hans"
    assert md["curve id"] == "hans"


def test_lazy_metadata():
    md = am.MetaData({"enum": "2",
                      "z range": am.LazyMetaValue(np.abs, -3),
                      "imaging mode": "force-distance"})
    value1 = super(am.MetaData, md).__getitem__("z range")
    assert isinstance(value1, am.LazyMetaValue)
    value2 = md["z range"]
    assert not isinstance(value2, am.LazyMetaValue)
    assert value2 == 3
    # the LazyMetaValue should have been replaced with the value
    value3 = super(am.MetaData, md).__getitem__("z range")
    assert not isinstance(value3, am.LazyMetaValue)
    assert value3 == 3
    assert md["segment count"] == 2


def test_lazy_metadata_copy():
    md = am.MetaData({"enum": "2",
                      "z range": am.LazyMetaValue(np.abs, -3),
                      "imaging mode": "force-distance"})
    mdc = md.copy()
    value1 = super(am.MetaData, md).__getitem__("z range")
    value2 = super(am.MetaData, mdc).__getitem__("z range")
    assert isinstance(value1, am.LazyMetaValue)
    assert isinstance(value2, am.LazyMetaValue)
    assert value1 is value2, "LazyMetaValue should not be copied"
    # access it and make sure that LazyMetaValue is overridden in md
    assert md["z range"] == 3
    value3 = super(am.MetaData, md).__getitem__("z range")
    assert not isinstance(value3, am.LazyMetaValue)
    # for mdc, this should not have happened
    value4 = super(am.MetaData, mdc).__getitem__("z range")
    assert isinstance(value4, am.LazyMetaValue)
    # but the value should be set
    assert value4.value == 3


def test_lazy_metadata_deepcopy():
    md = am.MetaData({"enum": "2",
                      "z range": am.LazyMetaValue(np.abs, -3),
                      "imaging mode": "force-distance"})
    mdc = copy.deepcopy(md)
    value1 = super(am.MetaData, md).__getitem__("z range")
    value2 = super(am.MetaData, mdc).__getitem__("z range")
    assert isinstance(value1, am.LazyMetaValue)
    assert isinstance(value2, am.LazyMetaValue)
    assert value1 is value2, "LazyMetaValue should not be copied"
    # access it and make sure that LazyMetaValue is overridden in md
    assert md["z range"] == 3
    value3 = super(am.MetaData, md).__getitem__("z range")
    assert not isinstance(value3, am.LazyMetaValue)
    # for mdc, this should not have happened
    value4 = super(am.MetaData, mdc).__getitem__("z range")
    assert isinstance(value4, am.LazyMetaValue)
    # but the value should be set
    assert value4.value == 3


def test_values_with_lazy_meta():
    md = am.MetaData({"enum": 2,
                      "z range": am.LazyMetaValue(np.abs, -3)})
    assert list(md.values()) == [2, 3]


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



import numpy as np

from .. import errors


__all__ = ["load_txt"]


converters = {
    0: lambda s: float(s.decode().strip().replace(",", ".")),
    1: lambda s: float(s.decode().strip().replace(",", ".")),
    2: lambda s: float(s.decode().strip().replace(",", ".")),
}


def crop_beginning(data):
    """Crop constant padding at the begin of the data column"""
    assert data.shape[1] == 2

    start = data[0, 1]
    for ii in range(1, data.shape[0]):
        if data[ii, 1] != start:
            break
    else:
        raise ValueError("Encountered all-constant data!")

    return data[ii-1:, :]


def detect_txt(path):
    """File should be plain text"""
    valid = False
    try:
        rawdata = np.loadtxt(path, converters=converters, max_rows=10)
    except (ValueError, IndexError):
        pass
    else:
        if np.issubdtype(rawdata.dtype, np.floating):
            valid = True
    return valid


def load_txt(path, callback=None, meta_override=None):
    """Load text files exported by the NT-MDT Nova software

    The columns are assumed to be: height (piezo) [nm],
    Deflection during approach (nA), Deflection during retraction (nA).
    The sensitivity in meta_override should be given in [m/A]
    (even though it is displayed as [m/V]).

    This loader removes constant-value padding at the beginning of
    the data columns, an artifact that is sometimes introduced during
    data export. There are no metadata in this file format.

    Test data were provided by Yuri Efremov
    :cite:`Efremov_figshare_20` :cite:`Efremov_2015`.

    Note that support for the original .mdt files is currently (2020)
    not possible. There exist binary readers for nt-mdt files
    (https://github.com/kaitai-io/kaitai_struct/), but this does
    not work for `exemplary data
    <https://doi.org/10.6084/m9.figshare.11862327.v1>`_. If the
    NT-MDT Nova software is not available, it should still be
    possible to load the data with `Ggyddion <http://gwyddion.net>`_
    and export it to something afmformats understands.

    Parameters
    ----------
    path: str or pathlib.Path or io.TextIOBase
        path to an ntmdt-exported .txt file
    callback: callable
        function for progress tracking; must accept a float in
        [0, 1] as an argument.
    meta_override: dict
        if specified, contains key-value pairs of metadata that
        are used when loading the files
        (see :data:`afmformats.meta.META_FIELDS`)
    """
    if meta_override is None:
        meta_override = {}
    req_metadata = ["sensitivity", "spring constant"]
    mis_metadata = [key for key in req_metadata if key not in meta_override]
    if mis_metadata:
        raise errors.MissingMetaDataError(
            mis_metadata,
            f"Please specify {' and '.join(mis_metadata)}!")

    rawdata = np.loadtxt(path, converters=converters)

    if rawdata.shape[1] != 3:
        raise errors.InvalidFileFormatError(
            "Expected 3 columns, got {}: {}".format(rawdata.shape[1], path))

    raw_apr = crop_beginning(rawdata[:, :2])
    raw_ret = crop_beginning(rawdata[:, ::2])

    data = {"height (measured)": np.concatenate((raw_apr[::-1, 0],
                                                 raw_ret[:, 0])) * 1e-9}
    fmult = meta_override["sensitivity"] * meta_override["spring constant"]
    data["force"] = np.concatenate((raw_apr[::-1, 1],
                                    raw_ret[:, 1])) * fmult * 1e-9
    data["segment"] = np.concatenate(
        (np.zeros(raw_apr.shape[0], dtype=np.uint8),
         np.ones(raw_ret.shape[0], dtype=np.uint8)))

    metadata = {"enum": 0,
                "imaging mode": "force-distance",
                "path": path,
                }
    metadata.update(meta_override)

    dd = {"data": data,
          "metadata": metadata}

    if callback is not None:
        callback(1)

    return [dd]


recipe_ntmdt_txt = {
    "descr": "exported by NT-MDT Nova",
    "detect": detect_txt,
    "loader": load_txt,
    "suffix": ".txt",
    "modalities": ["force-distance"],
    "maker": "NT-MDT Spectrum Instruments",
}



import numpy as np

from ...lazy_loader import LazyData
from ...meta import LazyMetaValue

from .jpk_reader import JPKReader
from .jpk_meta import ReadJPKMetaKeyError


__all__ = ["load_jpk"]


def detect(path):
    """Check whether a file is a valid JPK data file

    """
    # The suffix is not checked, because that is done in
    # the wrapper class formats.AFMFormatRecipe.
    jpkr = JPKReader(path)
    try:
        jpkr.get_metadata(index=0)
    except ReadJPKMetaKeyError:
        valid = False
    else:
        valid = True
    return valid


def load_jpk(path, callback=None, meta_override=None):
    """Loads JPK Instruments data files

    These files are zip files containing java property files and
    integer-encoded binary data. The property files include recipes
    on how to convert the raw integer data to SI units.

    Parameters
    ----------
    path: str or pathlib.Path
        path to JPK data file
    callback: callable
        function for progress tracking; must accept a float in
        [0, 1] as an argument.
    meta_override: dict
        if specified, contains key-value pairs of metadata that
        are used when loading the files
        (see :data:`afmformats.meta.META_FIELDS`)
    """
    if meta_override is None:
        meta_override = {}
    else:
        # just make sure nobody expects a different result for the forces
        for key in ["sensitivity", "spring constant"]:
            if key in meta_override:
                raise NotImplementedError(
                    f"Setting metadata such as '{key}' is not implemented!")

    jpkr = JPKReader(path)
    dataset = []
    # iterate over all datasets and add them
    for index in range(len(jpkr)):
        lazy_data = LazyData()
        for column in ["force", "height (measured)", "height (piezo)",
                       "segment", "time"]:
            lazy_data.set_lazy_loader(column=column,
                                      func=jpkr.get_data,
                                      kwargs={"column": column,
                                              "index": index})
        metadata = jpkr.get_metadata(index=index)
        metadata["z range"] = LazyMetaValue(
            lambda data: np.ptp(data["height (piezo)"]),
            lazy_data)
        metadata.update(meta_override)
        dataset.append({"data": lazy_data,
                        "metadata": metadata,
                        })
        if callback:
            callback((1+index) / len(jpkr))
    return dataset


recipe_jpk_force = {
    "descr": "binary FD data",
    "detect": detect,
    "loader": load_jpk,
    "maker": "JPK Instruments",
    "modalities": ["creep-compliance", "force-distance", "stress-relaxation"],
    "suffix": ".jpk-force",
}

recipe_jpk_force_map = {
    "descr": "binary QMap data",
    "detect": detect,
    "loader": load_jpk,
    "maker": "JPK Instruments",
    "modalities": ["creep-compliance", "force-distance", "stress-relaxation"],
    "suffix": ".jpk-force-map",
}

recipe_jpk_force_qi_data = {
    "descr": "binary QMap data",
    "detect": detect,
    "loader": load_jpk,
    "maker": "JPK Instruments",
    "modalities": ["creep-compliance", "force-distance", "stress-relaxation"],
    "suffix": ".jpk-qi-data",
}

recipe_jpk_force_qi_series = {
    "descr": "binary QMap data",
    "detect": detect,
    "loader": load_jpk,
    "maker": "JPK Instruments",
    "modalities": ["creep-compliance", "force-distance", "stress-relaxation"],
    "suffix": ".jpk-qi-series",
}



import functools
import pathlib

import h5py
import numpy as np

from ..afm_data import column_dtypes, known_columns
from ..meta import IMAGING_MODALITIES


__all__ = ["H5DictReader", "load_hdf5"]


class H5DictReader(object):
    def __init__(self, path_or_h5, enum_key):
        """Read-only HDF5-based dictionary for arrays

        Parameters
        ----------
        path_or_h5: str or pathlib.Path or h5py.Group
            Path to HDF5 file or an HDF5 group
        enum_key: str
            Name of the subgroup in `path_or_h5` that contains the data
            of the dictionary
        """
        if isinstance(path_or_h5, h5py.Group):
            # we are not responsible for the HDF5 file
            self.path = None
            self.h5 = path_or_h5
        else:
            # we are responsible for closing the HDF5 file
            self.path = path_or_h5
            self.h5 = None
        self.enum_key = enum_key
        self._columns = self.keys()

    def __contains__(self, key):
        return key in self._columns

    def __getitem__(self, key):
        if key not in known_columns:
            raise ValueError("Column '{}' is not documented!".format(key))
        elif key in self.keys():
            if self.path is not None:
                with h5py.File(self.path, "r") as h5:
                    val = np.asarray(h5[self.enum_key][key][:],
                                     dtype=column_dtypes[key])
            else:
                val = np.asarray(self.h5[self.enum_key][key][:],
                                 dtype=column_dtypes[key])
        else:
            raise KeyError("Column '{}' not in '{}/{}'".format(key, self.path,
                                                               self.enum_key))
        return val

    def __iter__(self):
        for kk in self._columns:
            yield kk

    @functools.lru_cache(maxsize=2)
    def keys(self):
        if self.path is not None:
            with h5py.File(self.path, "r") as h5:
                cols = sorted(h5[self.enum_key].keys())
        else:
            cols = sorted(self.h5[self.enum_key].keys())
        return cols


def detect_hdf5(path):
    """Detect HDF5 file format"""
    with h5py.File(path, mode="r") as h5:
        if "software" not in h5.attrs:
            return False
        elif "software version" not in h5.attrs:
            return False
        elif "0" not in h5:
            return False
        elif h5["0"].attrs["imaging mode"] not in IMAGING_MODALITIES:
            return False
        else:
            return True


def load_hdf5(path_or_h5, callback=None, meta_override=None):
    """Loads HDF5 files as exported by afmformats

    The HDF5 format is self explanatory. The root attributes
    contain the version of afmformats used to create it. For each
    curve, one group is created, named according to "0", "1", ...
    "9", "10", "11", etc. The attributes of each group are key-value
    pairs defined in :const:`afmformats.meta.KEYS_VALID`. The group
    contains datasets named according to
    :const:`afmformats.afm_data.known_columns` and have the attribute
    "unit" with the corresponding value in
    :const:`afmformats.afm_data.column_units`.

    Parameters
    ----------
    path_or_h5: str or pathlib.Path or h5py.Group
        path to HDF5 file or an HDF5 group
    callback: callable
        function for progress tracking; must accept a float in
        [0, 1] as an argument.
    meta_override: dict
        if specified, contains key-value pairs of metadata that
        are used when loading the files
        (see :data:`afmformats.meta.META_FIELDS`)

    Notes
    -----
    In case `path_or_h5` is a h5py.Group object, the
    "path" metadata variable will always be set to the
    path of the original HDF5 file. Keep this in mind
    if you think about storing multiple datasets (each
    containing multiple curves) in one HDF5 file (bad idea).
    """
    if meta_override is None:
        meta_override = {}
    else:
        # just make sure nobody expects a different result for the forces
        for key in ["sensitivity", "spring constant"]:
            if key in meta_override:
                raise NotImplementedError(
                    f"Setting metadata such as '{key}' is not implemented!")

    if isinstance(path_or_h5, h5py.Group):
        path = pathlib.Path(path_or_h5.file.filename)
        close = False
        h5 = path_or_h5
    else:
        path = pathlib.Path(path_or_h5)
        close = True
        h5 = h5py.File(path, "r")
    fdlist = []
    for enum_key in h5.keys():
        metadata = dict(h5[enum_key].attrs)
        metadata["path"] = path
        metadata["enum"] = int(enum_key)
        metadata.update(meta_override)
        data = H5DictReader(path_or_h5, enum_key=enum_key)
        fdlist.append({"data": data,
                       "metadata": metadata})
    if close:
        h5.close()
    if callback is not None:
        callback(1)
    return fdlist


recipe_hdf5 = {
    "descr": "HDF5-based",
    "detect": detect_hdf5,
    "loader": load_hdf5,
    "suffix": ".h5",
    "modalities": ["force-distance"],
    "maker": "afmformats",
}



# flake8: noqa: F401
from . import meta
from .afm_group import AFMGroup
from .afm_qmap import AFMQMap
from .formats import find_data, load_data
from .formats import supported_extensions
from .mod_creep_compliance import AFMCreepCompliance
from .mod_force_distance import AFMForceDistance
from .mod_stress_relaxation import AFMStressRelaxation

from ._version import version as __version__



import pathlib

import numpy as np

from .parse_funcs import fint, vd_str_in


__all__ = ["IMAGING_MODALITIES", "META_FIELDS", "DEF_ALL", "KEYS_VALID",
           "MetaDataMissingError", "LazyMetaValue", "MetaData", "parse_time"]


#: supported imaging modalities
IMAGING_MODALITIES = [
    # creep-compliance (constant force)
    "creep-compliance",
    # force-distance (indent until setpoint and retract)
    "force-distance",
    # stress-relaxation (constant indentation)
    "stress-relaxation",
    ]

#: Compendium of all allowed meta data keys, sorted by topic, and
#: including units and validation methods
META_FIELDS = {
    # AFM acquisition settings
    "acquisition": {
        "feedback mode": ["Feedback mode", "", vd_str_in([
            # JPK contact mode
            "contact",
            # From the NanoWizard User Manual (v. 4.2) sec. 5.7:
            # Force modulation mode is a mixture between contact mode and
            # intermittent mode and "can be thought of as a kind of contact
            # mode with an added vibration on the cantilever".
            "force-modulation",
        ])],
        "imaging mode": ["Imaging modality", "",
                         vd_str_in(IMAGING_MODALITIES)],
        "sensitivity": ["Sensitivity", "m/V", float],
        "spring constant": ["Cantilever spring constant", "N/m", float],
    },
    # dataset parameters
    "dataset": {
        "duration": ["Duration", "s", float],
        "duration approach": ["Duration of approach segment", "s", float],
        "duration retract": ["Duration of retract segment", "s", float],
        "enum": ["Dataset index within the experiment", "", fint],
        "point count": ["Size of the dataset in points", "", fint],
        "rate approach": ["Sampling rate of approach segment", "Hz", float],
        "rate retract": ["Sampling rate of retract segment", "Hz", float],
        "segment count": ["Number of segments", "", fint],
        "setpoint": ["Target indentation force", "N", float],
        "speed approach": ["Piezo speed of approach segment", "m/s", float],
        "speed retract": ["Piezo speed of retract segment", "m/s", float],
        "z range": ["Axial piezo range", "m", float],
    },
    # parameters specific for creep-compliance
    "dataset-mod creep-compliance": {
        "duration intermediate": ["Duration of intermediate segment",
                                  "s", float],
    },
    # QMap related dataset metadata
    "qmap": {
        "grid center x": ["Horizontal center of grid", "m", float],
        "grid center y": ["Vertical center of grid", "m", float],
        "grid index x": ["Horizontal grid position index", "", fint],
        "grid index y": ["Vertical grid position index", "", fint],
        "grid shape x": ["Horizontal grid shape", "px", fint],
        "grid shape y": ["Vertical grid shape", "px", fint],
        # The grid size is the actual size when you include the
        # boundary of the image (i.e. (grid_shape + 1) * pixel_size).
        "grid size x": ["Horizontal grid image size", "m", float],
        "grid size y": ["Vertical grid image size", "m", float],
        "position x": ["Horizontal position", "m", float],
        "position y": ["Vertical position", "m", float],
    },
    # AFM setup
    "setup": {
        "instrument": ["Instrument", "", str],
        "software": ["Acquisition software", "", str],
        "software version": ["Acquisition software version", "", str],
    },
    # storage data
    "storage": {
        "curve id": ["Curve identifier", "", str],
        "date": ["Recording date", "", str],  # YYYY-MM-DD
        "format": ["File format", "", str],
        "path": ["Path", "", pathlib.Path],
        "session id": ["Dataset identifier", "", str],
        "time": ["Recording time", "", str],  # HH:MM:SS.S
    },
}


#: A dictionary for all metadata definitions
DEF_ALL = {}
for _sec in META_FIELDS:
    for _key in META_FIELDS[_sec]:
        DEF_ALL[_key] = META_FIELDS[_sec][_key]

#: List of all valid meta data keys
KEYS_VALID = sorted(DEF_ALL.keys())


class MetaDataMissingError(BaseException):
    """Raised when meta data is missing"""
    pass


class LazyMetaValue:
    """A metadata value that is evaluated lazily in :class:`MetaData`"""
    def __init__(self, func, *args, **kwargs):
        """

        Example usage::

           meta = afmformats.meta.MetaData
           meta["z range"] = afmformats.meta.LazyMetaValue(
                np.ptp,
                np.arange(10))

        Parameters
        ----------
        func: callable
            Function to call to get the metadata value
        args:
            arguments to ``func``
        kwargs:
            Keyword arguments to ``func``
        """
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.value = None

    def __call__(self):
        if self.value is None:
            # When metadata is copied, LazyMetaValue is copied along with
            # it. If in the copy, this function here is called, then we
            # store it in self.value, so it does not have to be recomputed
            # in another copy.
            self.value = self.func(*self.args, **self.kwargs)
        return self.value


class MetaData(dict):
    """Management of meta data variables

    Valid key names are defined in :const:`afmformats.meta.KEYS_VALID`.
    """
    valid_keys = KEYS_VALID

    def __init__(self, *args, **kwargs):
        # do not init with args/kwargs
        super(MetaData, self).__init__()
        # instead, make sure everything goes through __set__
        self.update(*args, **kwargs)
        # check for invalid keys
        for key in self:
            if key not in self.valid_keys:
                raise KeyError("Unknown metadata key: '{}'".format(key))

    def __copy__(self):
        """Create a copy of the metadata

        Returns
        -------
        mdc: MetaData
            Copy of the MetaData class (LazyMetaValue not copied)
        """
        cls = MetaData()
        for key in self:
            # This will just pass the LazyMetaValue instance and
            # not create a copy of it.
            cls[key] = super(MetaData, self).__getitem__(key)
        return cls

    def __deepcopy__(self, memo):
        # Since MetaData is not a nested dictionary, we just do a copy.
        # And LazyMetaValue should not be copied.
        result = self.__copy__()
        memo[id(self)] = result
        return result

    def __setitem__(self, key, value):
        """Set a metadata key

        The key must be a valid key defined in `self.valid_keys`
        (defaults to :const:`afmformats.meta.KEYS_VALID`).

        The "time" key is converted using :func:`parse_time`.
        NaN values are silently ignored.
        """
        if key not in self.valid_keys:
            raise KeyError("Unknown metadata key: '{}'".format(key))
        elif key == "time":
            value = parse_time(value)
        elif key == "imaging mode" and "segment count" not in self:
            if value == "force-distance":
                self["segment count"] = 2
            elif value in ["creep-compliance", "stress-relaxation"]:
                self["segment count"] = 3
            else:
                # Dear future self...
                raise ValueError(f"Please add '{value}' to this case!")
        if isinstance(value, float) and np.isnan(value):
            # nan values are ignored
            return
        if not isinstance(value, LazyMetaValue):
            # parse the value
            value = DEF_ALL[key][2](value)
        super(MetaData, self).__setitem__(key, value)
        self._autocomplete_grid_metadata()

    def __getitem__(self, key):
        if key == "curve id":
            return self._get_curve_id()
        elif key == "session id":
            return self._get_session_id()
        elif key not in self and key in self.valid_keys:
            msg = "No meta data was defined for '{}'! ".format(key) \
                  + "Please make sure you passed the dictionary `metadata` " \
                  + "when you loaded your data."
            raise MetaDataMissingError(msg)
        elif key not in self:
            msg = "Unknown meta key: '{}'!".format(key)
            raise KeyError(msg)

        value = super(MetaData, self).__getitem__(key)

        if isinstance(value, LazyMetaValue):
            # Evaluate LazyMetaValue and assign value to self
            value = value()
            # parse the value
            value = DEF_ALL[key][2](value)
            super(MetaData, self).__setitem__(key, value)
        return value

    def _autocomplete_grid_metadata(self):
        """Compute missing grid metadata in-place

        - supplements "grid index x" and "grid index y" if possible
        """
        md = self
        for ax in ["x", "y"]:
            if (f"grid index {ax}" not in self
                and f"grid center {ax}" in self
                and f"grid shape {ax}" in self
                and f"grid size {ax}" in self
                    and f"position {ax}" in self):
                idx = self._convert_position_to_index(
                    pos_m=md[f"position {ax}"],
                    size_m=md[f"grid size {ax}"],
                    center_m=md[f"grid center {ax}"],
                    size_px=md[f"grid shape {ax}"])
                super(MetaData, self).__setitem__(f"grid index {ax}", idx)

    @staticmethod
    def _convert_position_to_index(pos_m, size_m, center_m, size_px):
        """Convert qmap positions from [m] to array coordinates in [px]

        Parameters
        ----------
        pos_m: float
            positions [m]
        size_m: float
            grid size [m]
        center_m: float
            grid center position [m]
        size_px: int
            grid size [px]

        Returns
        -------
        pos_px: int
            index position of `pos_m`
        """
        if size_px != int(size_px):
            raise ValueError(
                "`size_px` must be integer, got {}!".format(size_px))
        size_px = int(size_px)
        s1 = center_m - size_m / 2
        s2 = center_m + size_m / 2

        x, dx = np.linspace(s1, s2, size_px, endpoint=False, retstep=True)
        x += dx / 2

        xpx = np.nanargmin(np.abs(x - pos_m))
        return xpx

    def _get_curve_id(self):
        # already set?
        thisid = super(MetaData, self).get("curve id")
        # compute using session_id
        if not thisid and "enum" in self:
            thisid = self._get_session_id() + "_{}".format(self["enum"])
        # not available
        if not thisid:
            raise MetaDataMissingError("Key 'curve id' not set!")
        return thisid

    def _get_session_id(self):
        # already set?
        thisid = super(MetaData, self).get("session id")
        # compute using date/time
        if not thisid:
            idlist = [self.get("date", ""),
                      self.get("time", "")]
            thisid = "_".join([it for it in idlist if it])
        # not available
        if not thisid:
            raise MetaDataMissingError("Key 'session id' not set!")
        return thisid

    def as_dict(self):
        """Convert to real dictionary

        This is needed e.g. for `self.items` such that `json.dump`
        works in combination with `LazyMetaValue` (which is not
        JSON serializable)
        """
        realdict = {}
        for key in self:
            realdict[key] = self[key]
        return realdict

    def copy(self):
        """Create a copy of the metadata

        Returns
        -------
        mdc: MetaData
            Copy of the MetaData class (LazyMetaValue not copied)
        """
        return self.__copy__()

    def get(self, key, default=None):
        if key in self:
            return self[key]
        else:
            return default

    def get_summary(self):
        """Convenience function returning the meta data summary

        Returns a dict of dicts with keys matching the DEF_* dicts.
        Unset values are returned as `np.nan`.
        """
        summary = {}
        for sec in META_FIELDS:
            if sec.startswith("dataset-mod"):
                if self["imaging mode"] not in sec:
                    # Ignore metadata that do not belong to the current
                    # imaging modality.
                    continue
                else:
                    # Write "dataset-mod" metadata to the "dataset" section
                    sum_sec = "dataset"
            else:
                sum_sec = sec
            summary.setdefault(sum_sec, {})
            for key in META_FIELDS[sec]:
                summary[sum_sec][key] = self.get(key, np.nan)
        return summary

    def items(self):
        return self.as_dict().items()

    def update(self, *args, **kwargs):
        # make sure everything goes through __set__
        for k, v in dict(*args, **kwargs).items():
            self[k] = v

    def values(self):
        return self.as_dict().values()


def parse_time(value):
    """Convert a time string to "HH:MM:SS.S"

    - leading zeros are added where necessary
    - trailing zeros after "." are stripped
    - trailing "." is stripped

    e.g.

    - "6:15:22 PM" -> "18:15:22"
    - "6:15:22.00 AM" -> "06:15:22"
    - "6:02:22.010 AM" -> "06:02:22.01"
    """
    # fix AM/PM
    if value.count(" "):
        time, ap = value.split()
        if ap == "PM":  # convert to 24h
            timetup = time.split(":")
            timetup[0] = str(int(timetup[0]) + 12)
            time = ":".join(timetup)
    else:
        time = value
    # convert to time with leading zeros and stripped subseconds
    hh, mm, ss = time.split(":")
    if ss.count("."):
        ss0, ss1 = ss.split(".")
        ss1 = ("." + ss1).rstrip(".0")
    else:
        ss0 = ss
        ss1 = ""
    newvalue = ":".join(["{:02d}".format(int(hh)),
                         "{:02d}".format(int(mm)),
                         "{:02d}".format(int(ss0)) + ss1
                         ])
    return newvalue



import pathlib

import pytest

import afmformats
import afmformats.errors


data_path = pathlib.Path(__file__).parent / "data"


@pytest.mark.parametrize("path", data_path.glob("fmt-*-fd_*"))
def test_load_force_distance_with_callback(path):
    """Make sure that the callback function is properly implemented"""
    calls = []

    def callback(value):
        calls.append(value)

    try:
        afmformats.load_data(path=path, callback=callback)
    except afmformats.errors.MissingMetaDataError:
        afmformats.load_data(path=path,
                             callback=callback,
                             meta_override={"spring constant": 20,
                                            "sensitivity": .01e-6})
    assert calls[-1] == 1


@pytest.mark.parametrize("path", data_path.glob("fmt-*-cc_*"))
def test_load_creep_compliance_with_callback(path):
    """Make sure that the callback function is properly implemented"""
    calls = []

    def callback(value):
        calls.append(value)

    try:
        afmd = afmformats.load_data(path=path, callback=callback)
    except afmformats.errors.MissingMetaDataError:
        afmd = afmformats.load_data(path=path,
                                    callback=callback,
                                    meta_override={"spring constant": 20,
                                                   "sensitivity": .01e-6})
    for cc in afmd:
        assert cc.modality == "creep-compliance"
    assert calls[-1] == 1



"""Test of basic opening functionalities"""
import pathlib
import shutil
import tempfile

import numpy as np
import pytest

import afmformats
import afmformats.errors

from afmformats.formats.fmt_jpk.jpk_reader import ArchiveCache


data_path = pathlib.Path(__file__).resolve().parent / "data"


def test_archive_cache():
    td = pathlib.Path(tempfile.mkdtemp(prefix="archive_cache_jpk_"))
    ziplist = []
    for ii in range(ArchiveCache.max_archives + 1):
        pnew = td / f"spot_{ii:03d}.jpk-force"
        shutil.copy2(data_path / "fmt-jpk-fd_spot3-0192.jpk-force", pnew)
        ziplist.append(ArchiveCache.get(pnew))
    assert ziplist[0].fp is None  # we are one over `max_archives`
    assert not ziplist[1].fp is None
    assert not ziplist[-1].fp is None


def test_creep_compliance1():
    jpkfile = data_path / "fmt-jpk-cc_pr14-brain-2021.06.30.jpk-force"
    ds = afmformats.load_data(path=jpkfile)[0]
    assert np.allclose(ds.metadata["duration intermediate"], 3,
                       atol=0, rtol=1e-12)


def test_creep_compliance2():
    jpkfile = data_path / "fmt-jpk-cc_pr14-polymergel-2021.06.30.jpk-force"
    ds = afmformats.load_data(path=jpkfile)[0]
    assert np.allclose(ds.metadata["duration intermediate"], 3,
                       atol=0, rtol=1e-12)


@pytest.mark.parametrize("name, is_valid",
    [("fmt-jpk-fd_spot3-0192.jpk-force", True),  # noqa: E128
     ("fmt-jpk-fd_map2x2_extracted.jpk-force-map", True),
     ("fmt-jpk-cl_calibration_force-save-2015.02.04-11.25.21.294.jpk-force",
      False),
     ])
def test_detect_jpk(name, is_valid):
    jpkfile = data_path / name
    if is_valid:
        afmlist = afmformats.load_data(path=jpkfile)
        assert afmlist
    else:
        with pytest.raises(afmformats.errors.FileFormatNotSupportedError):
            afmformats.load_data(path=jpkfile)


def test_get_z_range():
    jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
    ds = afmformats.load_data(path=jpkfile)[0]
    assert np.allclose(ds.metadata["z range"], 4.280850000996709e-06,
                       atol=0, rtol=1e-6)


def test_load_jpk_map():
    jpkfile = data_path / "fmt-jpk-fd_map2x2_extracted.jpk-force-map"
    afmlist = afmformats.load_data(path=jpkfile)

    assert len(afmlist) == 4
    ds = afmlist[2]
    ds2 = afmlist[3]
    assert ds.metadata["enum"] == 2
    assert np.allclose(ds.metadata["duration"], 1.6089775561097257)
    # Verified with visual inspection of force curve in JPK software
    assert np.allclose(ds["force"][0], -5.8540192943834714e-10)
    assert np.allclose(ds["height (measured)"][0], 0.0001001727719556085)
    # make sure curve id is different
    assert ds.metadata["curve id"] != ds2.metadata["curve id"]
    assert ds.metadata["session id"] == ds2.metadata["session id"]


def test_load_jpk_simple():
    jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
    afmlist = afmformats.load_data(path=jpkfile)
    ds = afmlist[0]
    assert ds.metadata["enum"] == 0
    assert np.allclose(ds["height (measured)"][0], 2.2815672438768612e-05)


def test_load_jpk_piezo():
    jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
    afmlist = afmformats.load_data(path=jpkfile)
    ds = afmlist[0]
    assert np.allclose(ds["height (piezo)"][0], 2.878322343068329e-05)


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



import pathlib
from unittest import mock

import pytest

import numpy as np

from afmformats import AFMGroup, AFMQMap

data_path = pathlib.Path(__file__).parent / "data"


def test_feat_scan_order():
    qm = AFMQMap(data_path / "fmt-jpk-fd_map2x2_extracted.jpk-force-map")
    order = qm.get_qmap("data: scan order", qmap_only=True)
    assert order[0, 0] == 0
    assert order[0, -1] == 1
    assert order[-1, -1] == 2
    assert order[-1, 0] == 3
    assert np.isnan(order[0, 1])


def test_feat_min_height():
    qm = AFMQMap(data_path / "fmt-jpk-fd_map2x2_extracted.jpk-force-map")
    qd = qm.get_qmap("data: height base point", qmap_only=True)
    assert np.allclose(qd[0, 0], 40.55030392499141)
    assert np.allclose(qd[0, -1], 47.354988549298945)
    assert np.allclose(qd[-1, -1], 96.1627883099352)
    assert np.allclose(qd[-1, 0], 89.95170867840217)


def test_get_coords():
    qm = AFMQMap(data_path / "fmt-jpk-fd_map2x2_extracted.jpk-force-map")

    px = qm.get_coords(which="px")
    refpx = np.array([[0, 0], [9, 0], [9, 9], [0, 9]])
    assert np.all(px == refpx)

    um = qm.get_coords(which="um")
    refum = np.array([[31.972656250000004, -753.5351562500001],
                      [571.8359375000001, -753.90625],
                      [571.8359375000001, -213.73046875000003],
                      [31.855468750000004, -213.73046875000003]])
    assert np.all(um == refum)


def test_get_coords_bad():
    qm = AFMQMap(data_path / "fmt-jpk-fd_map2x2_extracted.jpk-force-map")
    try:
        qm.get_coords(which="mm")
    except ValueError:
        pass
    else:
        assert False, "Units [mm] should not be supported."


def test_get_qmap():
    qm = AFMQMap(data_path / "fmt-jpk-fd_map2x2_extracted.jpk-force-map")
    x, y, _ = qm.get_qmap(feature="data: height base point", qmap_only=False)
    assert x.size == 10
    assert y.size == 10


def test_init_with_group():
    group = AFMGroup(data_path / "fmt-jpk-fd_map2x2_extracted.jpk-force-map")
    qm = AFMQMap(group)
    assert qm.shape == (10, 10)


def test_init_with_group_meta_override_meaningless():
    group = AFMGroup(data_path / "fmt-jpk-fd_map2x2_extracted.jpk-force-map")
    with pytest.raises(ValueError, match="meaningless"):
        AFMQMap(group,
                meta_override={"spring constant": 1,
                               "sensitivity": 2})


def test_metadata_missing():
    fake_qmap_dict = {
        "test-key": ["A description for the test key", "m", float],
    }
    group = AFMGroup(data_path / "fmt-jpk-fd_map2x2_extracted.jpk-force-map")
    with mock.patch("afmformats.afm_qmap.META_FIELDS",
                    {"qmap": fake_qmap_dict}):
        with pytest.raises(ValueError,
                           match="QMap metadata missing .* test-key"):
            AFMQMap(group)


def test_shape():
    qm = AFMQMap(data_path / "fmt-jpk-fd_map2x2_extracted.jpk-force-map")
    assert np.allclose(qm.extent,
                       [1.97265625, 601.97265625,
                        -783.53515625, -183.53515625000006])
    assert qm.shape == (10, 10)



import numpy as np

__all__ = ["AFMSegment"]


class AFMSegment(object):
    """Simple wrapper around dict-like `data` to expose a single segment

    This class also caches the segment indices.
    """
    def __init__(self, raw_data, data, segment):
        """New Segment data

        Parameters
        ----------
        raw_data: dict
            dictionary containing valid column names as keys and
            1d ndarrays as values; this is raw data (e.g. from the
            measurement file) that may be lazily-loaded
        data: dict
            same as raw_data, but in this case the data are already
            in memory; we distinguish between raw_data and data so
            that we know where the data came from (e.g. there might
            be "tip poisition" in both dictionaries, but we only
            always use (and override) the "tip position" in `data`.
            We never touch `raw_data`.
        """
        #: The segment type (approach, intermediate, or retract)
        self.segment = segment
        self._raw_data = raw_data
        self._data = data

        self._raw_segment_indices = None
        self._user_segment_indices = None

    def __getitem__(self, key):
        """Access column data of the segment"""
        if key in self._data:
            return self._data[key][self.segment_indices]
        elif key in self._raw_data:
            return self._raw_data[key][self.segment_indices].copy()
        else:
            raise KeyError("Undefined column '{}'!".format(key))

    def __setitem__(self, key, data):
        """Set column data of the segment"""
        if key not in self._data and key not in self._raw_data:
            raise KeyError("Undefined column '{}'!".format(key))
        elif key not in self._data:
            self._data[key] = np.array(self._raw_data[key], copy=True)
        self._data[key][self.segment_indices] = data

    @property
    def segment_indices(self):
        """boolean array of segment indices"""
        if "segment" in self._data:  # data takes precedence (user-edited)
            if self._user_segment_indices is None:
                self._user_segment_indices = \
                    self._data["segment"] == self.segment
            indices = self._user_segment_indices
        elif "segment" in self._raw_data:
            # indices from raw data can safely be cached (will not change)
            if self._raw_segment_indices is None:
                self._raw_segment_indices = \
                    self._raw_data["segment"] == self.segment
            indices = self._raw_segment_indices
        else:
            raise ValueError("Could not identify segment data!")
        return indices

    def clear_cache(self):
        """Invalidates the segment indices corresponding to `self.data`"""
        self._user_segment_indices = None



"""Test Asylum .ibw format"""
import pathlib

import numpy as np

import afmformats


data_path = pathlib.Path(__file__).resolve().parent / "data"


def test_open_with_metadata():
    jpkfile = data_path / "fmt-igor-fd_U3_3_p10004.ibw"
    fdat = afmformats.load_data(jpkfile, modality="force-distance")[0]
    for col in ['force', 'height (measured)', 'height (piezo)', 'segment']:
        assert col in fdat
    assert fdat.metadata["sensitivity"] == 2.0364e-07
    assert fdat.metadata["spring constant"] == 0.31451
    assert fdat.metadata["setpoint"] == 1.8e-07
    assert np.allclose(fdat["force"].max(), 1.778407e-07, rtol=1e-7, atol=0)
    assert fdat.metadata["point count"] == 10097
    assert fdat.metadata["time"] == '18:18:49'
    assert np.allclose(fdat["height (measured)"].min(), -1.0322237e-06,
                       rtol=1e-7, atol=0)


def test_open_issue_8():
    # test file provided by María Tenorio (CC0)
    jpkfile = data_path / "fmt-igor-fd_SiN_FD_plot.ibw"
    fdat = afmformats.load_data(jpkfile, modality="force-distance")[0]
    assert fdat.metadata["time"] == "10:47:31"


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()
